( *** S-Forth -- System Vocabulary Version 1.4 *** )



( ----------------------------------------------------- )
( * S-Forth ist ein PC-Projekt von DOS International *  )
( Beschreibung siehe DOS International Ausgabe 9/87 ff  )
( [c] 1987 Volker Everts, Redaktion DOS International   )
(                         Elfenstr. 40, 8000 Mnchen 83 )
( ----------------------------------------------------- )



vocabulary forth     ( System-Vocabulary )



( *** System-Variablen *** )

variable r0 $0000  ( Return-Stack )    ( ** ss% und ds% fortgefallen )
variable s0 $F800  ( Daten-Stack )
variable dp $4000  ( freier Speicher )
variable base 10   ( Zahlenbasis )
variable hld 0     ( Zeiger auf Ziffer )
variable dpl 0     ( Zeiger auf Dezimalpunkt )
variable out 0     ( Zeichenzhler fr EMIT )
variable in 0      ( Zeichenzhler Eingabe )
variable tib $F000 ( Terminal Input Buffer )     ( ** ax, bx, cx, dx fortgefallen )
variable leading 0 ( enthlt fhrende Null oder Space )
variable adr 0     ( String-Adresse )
variable ml 0      ( maximale String-Lnge )



( *** System-Konstanten *** )

constant dictionary $108  ( Start Dictionary-Link )  ( ** neu 1.4 )
constant 0 0     ( Null )
constant 1 1     ( Eins )
constant 2 2     ( Zwei )
constant bl 32   ( Leerzeichen )
constant bs 8    ( Backspace )
constant true -1 ( True-Flag )
constant false 0 ( False-Flag )



( *** Stack-Manipulation *** )

macro dup ( n --> n n )
  [ $4E ]     ( dec si      )
  [ $4E ]     ( dec si      )
  [ $89 $04 ] ( mov [si],ax )
;


macro drop ( n --> )
  [ $AD ] ( lodsw )
;


macro swap ( n1 n2 --> n2 n1 )
  [ $87 $04 ] ( xchg ax,[si] )
;


macro >r [ $50 ] ( push ax ) drop ;


macro r> dup [ $58 ] ( pop ax ) ;


macro r r> [ $50 ] ( push ax ) ;


: over ( n1 n2 --> n1 n2 n1 )
  dup [ $8B $44 $02 ] ( mov ax,[si+2] )
;


: -dup ( n --> n ? )
  dup if dup endif
;


: rot ( n1 n2 n3 --> n2 n3 n1 )
  >r swap r> swap
;


: ddup ( d --> d d )
  over over
;



( *** einfache Arithmetik *** )

macro + ( n1 n2 --> Summe )
  [ $01 $04 ] ( add  [si],ax )
  drop
;


macro - ( n1 n2 --> Differenz )
  [ $29 $04 ] ( sub  [si],ax )
  drop
;


macro * ( n1 n2 --> Produkt )
  [ $F7 $2C ] ( imul ax,[si] )
  [ $46 $46 ] ( inc si; inc si )
;


: / ( n1 n2 --> Quotient )
  [ $51 ]     ( push cx    )
  [ $89 $C1 ] ( mov  cx,ax )
  [ $AD ]     ( lodsw      )
  [ $33 $D2 ] ( xor  dx,dx )
  [ $F7 $F9 ] ( idiv cx    )
  [ $59 ]     ( pop  cx    )
;


: mod ( n1 n2 --> Rest )
  / [ $89 $D0 ] ( mov ax,dx )
;


: /mod ( n1 n2 --> Rest Quotient )
  /
  [ $4E ]     ( dec si      )
  [ $4E ]     ( dec si      )
  [ $89 $14 ] ( mov [si],dx )
;


: */ ( n1 n2 n3 --> n1*n2/n3 )
  [ $51 ]     ( push cx      )
  [ $89 $C1 ] ( mov  cx,ax   )
  [ $AD ]     ( lodsw        )
  [ $F7 $2C ] ( imul ax,[si] )
  [ $F7 $F9 ] ( idiv cx      )
  [ $59 ]     ( pop  cx      )
;


: */mod ( n1 n2 n3 --> Rest n1*n2/n3 )
  */
  [ $4E ]     ( dec si      )
  [ $4E ]     ( dec si      )
  [ $89 $14 ] ( mov [si],dx )
;


macro and ( n1 n2 --> and )
  [ $21 $04 ] ( and [si],ax ) ( ** gendert )
  drop
;


macro or ( n1 n2 --> or )
  [ $09 $04 ] ( or [si],ax ) ( ** gendert )
  drop
;


macro xor ( n1 n2 --> xor )
  [ $31 $04 ] ( xor [si],ax )  ( ** gendert )
  drop
;


macro minus ( n --> -n )
  [ $F7 $D8 ] ( neg  ax )
;


macro not ( n --> not n )
  [ $F7 $D0 ] ( not  ax )
;


macro 1+ ( n --> n+1 )
  [ $40 ]  ( inc ax )
;


macro 2+ ( n --> n+2 )
  1+ 1+
;


macro 1- ( n --> n-1 )
  [ $48 ]  ( dec ax )
;


macro 2- ( n --> n-2 )
  1- 1-
;



( *** Double, Unsigned, Mixed Arithmetik *** )


: d+ ( d1 d2 --> summe )
  [ $8B $14   ] ( mov dx,[si]   )
  [ $01 $54 4 ] ( add [si+4],dx )
  [ $11 $44 2 ] ( adc [si+2],ax )
  drop drop
;


: u* ( u1 u2 --> ud )
  [ $F7 $24 ] ( mul ax,[si] )
  [ $89 $04 ] ( mov [si],ax )
  [ $89 $D0 ] ( mov ax,dx   )
;


: u/ ( ud u --> Rest Quotient )
  [ $51 ]     ( push cx    )
  [ $89 $C1 ] ( mov  cx,ax )
  [ $8B $14 ] ( mov dx,[si] )
  [ $AD ]     ( lodsw       )
  [ $AD ]     ( lodsw       )
  [ $F7 $F1 ] ( div cx      )
  [ $59 ]     ( pop  cx     )
  [ $4E ]     ( dec si      )
  [ $4E ]     ( dec si      )
  [ $89 $14 ] ( mov [si],dx )
;


: m* ( n1 n2 --> d )
  [ $F7 $2C ] ( imul ax,[si] )
  [ $89 $04 ] ( mov  [si],ax )
  [ $89 $D0 ] ( mov  ax,dx   )
;


: m/ ( d n --> Rest Quotient )
  [ $51 ]     ( push cx     )
  [ $89 $C1 ] ( mov  cx,ax  )
  [ $8B $14 ] ( mov dx,[si] )
  [ $AD ]     ( lodsw       )
  [ $AD ]     ( lodsw       )
  [ $F7 $F9 ] ( idiv cx     )
  [ $59 ]     ( pop  cx     )
  [ $4E ]     ( dec si      )
  [ $4E ]     ( dec si      )
  [ $89 $14 ] ( mov [si],dx )
;


: m/mod ( ud u --> Rest dQuotient )
  dup >r 0 swap u/ rot rot r> u/ rot
;



( *** Speicher-Befehle *** )

macro @ ( adr --> n )
  [ $89 $C3 ] ( mov bx,ax   )
  [ $8B $07 ] ( mov ax,[bx] )
;


: c@ ( adr --> c )
  [ $89 $C3 ] ( mov  bx,ax   )
  [ $8A $07 ] ( mov  al,[bx] )
  [ $B4 $00 ] ( mov  ah,0    )
;


: ! ( n adr --> )
  [ $89 $C3 ] ( mov  bx,ax   )
  drop
  [ $89 $07 ] ( mov  [bx],ax )
  drop
;


: c! ( c adr --> )
  [ $89 $C3 ] ( mov  bx,ax   )
  drop
  [ $88 $07 ] ( mov  [bx],al )
  drop
;


: +! ( n adr --> )
  [ $89 $C3 ] ( mov  bx,ax   )
  drop
  [ $01 $07 ] ( add  [bx],ax )
  drop
;


: ++ ( Inkrement Adresse )
  [ $89 $C3 ] ( mov  bx,ax )
  drop
  [ $FF $07 ] ( inc  [bx]  )
;


: -- ( Dekrement Adresse )
  [ $89 $C3 ] ( mov  bx,ax )
  drop
  [ $FF $0F ] ( dec  [bx]  )
;


: cmove ( von bis n --> )
  [ $51 ]       ( push cx    )
  [ $89 $C1 ]   ( mov  cx,ax )
  drop
  [ $57 ]       ( push di    )
  [ $89 $C7 ]   ( mov  di,ax )
  drop
  [ $56 ]       ( push si    )
  [ $89 $C6 ]   ( mov  si,ax )
  [ $39 $F7 ]   ( cmp  di,si )
  [ $7C 7   ]   ( jl   +7    )
  [ $03 $F1 ]   ( add  si,cx )
  [ $03 $F9 ]   ( add  di,cx )
  [ $4E ]       ( dec  si    )
  [ $4F ]       ( dec  di    )
  [ $FD ]       ( std        )
  [ $F3 $A4 ]   ( rep  movsb )
  [ $FC ]       ( cld        )
  [ $5E ]       ( pop  si    )
  [ $5F ]       ( pop  di    )
  [ $59 ]       ( pop  cx    )
  drop
;


: fill ( adr n c --> )
  [ $89 $C2 ]   ( mov  dx,ax     )
  drop
  [ $51 ]       ( push cx        )
  [ $89 $C1 ]   ( mov  cx,ax     )
  drop
  [ $57 ]       ( push di        )
  [ $89 $C7 ]   ( mov  di,ax     )
  [ $89 $D0 ]   ( mov  ax,dx     )
  [ $F3 $AA ]   ( rep  stosb     )
  [ $5F ]       ( pop  di        )
  [ $59 ]       ( pop  cx        )
  drop
;


: erase 0 fill ;


: blanks bl fill ;


: toggle over c@ xor swap c! ;


: here dp @ ;


: pad here $44 + ;


: allot dp +! ;


: , here ! 2 allot ;


: c, here c! 1 allot ;


macro sp@
  dup
  [ $89 $F0 ] ( mov ax,si )
;


macro sp!
  [ $89 $C6 ] ( mov si,ax )
  [ $33 $C0 ] ( xor ax,ax )
;


macro rp@
  dup
  [ $89 $E0 ] ( mov ax,sp )
;


macro rp!
  [ $89 $C4 ] ( mov sp,ax )
  drop
;


: pick ( n1 n2...ni nj nk... nn j --> n1 n2...ni nj nk...nn nj )
  dup + sp@ + @    ( ** neu )
;


: roll ( n1 n2...ni nj nk...nn j --> n1 n2...ni nk...nn nj )
  dup >r pick       ( Stackelement herauspicken )
  dup sp@ dup 2+    ( Stackpointer-Adresse holen )
  r> 1+ dup + cmove ( Stackelemente bewegen )
  drop drop         ( ** neu )
;


: d@ ( adr --> d )          ( ** neu 1.4 )
  dup @      ( Low Word )
  swap 2+ @  ( High Word )
;


: d! ( d adr --> )          ( ** neu 1.4 )
  rot over !  ( Low Word )
  2+ !        ( High Word )
;



( *** Untersttzung strukturierter Befehle *** )

macro i ( Schleifenindex nach TOS kopieren )
  dup
  [ $89 $C8 ] ( mov ax,cx )
;


macro j ( zweiter Schleifenindex )
  r
;


macro leave ( Schleife verlassen )
  [ $89 $CF ] ( mov di,cx )
;


macro exit ( Wort abbrechen )
  [ $C3 ] ( ret )
;



( *** Vergleichsoperatoren *** )

: 0=
  [ $3D 0 0 ] ( cmp  ax,0 )
  [ $B8 0 0 ] ( mov  ax,0 )
  [ $75 1 ]   ( jnz  +1   )
  [ $48 ]     ( dec  ax   )
;


: 0<
  [ $3D 0 0 ] ( cmp  ax,0 )
  [ $B8 0 0 ] ( mov  ax,0 )
  [ $7D 1 ]   ( jge  +1   )
  [ $48 ]     ( dec  ax   )
;


: 0<> 0= not ;  ( ** neu 1.4 )


: = - 0= ;


: < - 0< ;


: > swap - 0< ;


: <> = not ;  ( ** neu 1.4 )


( *** Funktionen *** )

: abs ( n --> abs )
  dup 0< if minus endif
;


: sgn ( n --> signum )
  dup if dup abs / endif
;


: +- ( n1 n2 --> sgn*n1 )
  0< if abs minus else abs endif
;


: dminus ( d --> -d )
  not swap not swap 1 0 d+
;


: dabs ( d --> ud )
  dup 0< if dminus endif
;


: s->d ( n --> d )
  dup 0< if -1 else 0 endif
;


: d+- ( d n --> sgn*d )
  0< if dabs dminus else dabs endif
;




( *** einfache Ein-/Ausgabe *** )


macro (vector)  ( ** neu 1.4 )
  [ $E9 0 0 ]  ( jmp +0 )
;


: emit ( TOS als ASCII-Zeichen ausgeben )
  (vector)   ( ** eingefgt )
  [ $88 $C2 ] ( mov  dl,al )
  [ $B4 $02 ] ( mov  ah,2  )
  [ $CD $21 ] ( int  21H   )
  drop out ++
;


: inkey ( Tastatureingabe --> TOS; mit Echo )
  dup
  [ $B4 $01 ] ( mov  ah,1 )
  [ $CD $21 ] ( int  21H  )
  [ $B4 $00 ] ( mov  ah,0 )
;


: key ( Tastatureingabe --> TOS )
  (vector)  ( ** eingefgt )
  dup
  [ $B4 $08 ] ( mov  ah,8 )
  [ $CD $21 ] ( int  21H  )
  [ $B4 $00 ] ( mov  ah,0 )
;


: ?key ( Taste gedrckt ? )
  (vector)  ( ** eingefgt )
  dup
  [ $B4 $0B ] ( mov  ah,11 )
  [ $CD $21 ] ( int  21H   )
  [ $98     ] ( cbw        )
;


: cr ( Zeilenvorschub )
  13 emit 10 emit
;


: space ( Leerzeichen ausgeben )
  bl emit
;


: spaces ( n Leerzeichen ausgeben )
  1 do space loop
;


: count ( adr1 --> adr2 n )
  dup c@        ( Count-Byte holen )
  swap 1+ swap  ( Adresse erhhen )
;


: type ( adr count --> )
  over + 1- swap  ( Endadresse berechnen )
  do
    i c@ emit     ( Zeichen ausgeben )
  loop
;


: " ( --> adr ) ( Adresse von String-Konstante nach TOS ) ( ** Neu )
  r>             ( Adresse von Return-Stack holen )
  dup dup c@     ( Count-Byte holen )
  + 1+           ( Rckkehr-Adresse berechnen... )
  >r             ( ...und auf Return-Stack legen )
;


: ." ( Text ausgeben )
  r>          ( Adresse von Return-Stack holen )
  dup dup c@  ( Count-Byte holen               )
  + 1+        ( Rckkehr-Adresse berechnen ... )
  >r          ( ... und wieder auf Stack legen )
  count type  ( String ausgeben                )
;


: (bs)  ( Backspace bei Eingabestring durchfhren )
  adr @ c@ dup     ( Count-Byte holen )
  if
    1- adr @ c!    ( Count-Byte dekrementieren )
    emit space bs  ( Backspace und Space ausgeben )
  else
    drop drop 7    ( BEL auf Stack legen, falls Count = 0 )
  endif
;


: (app)  ( Zeichen an Eingabestring anfgen )
  adr @ c@          ( Count-Byte holen )
  dup ml c@ =       ( maximale Lnge erreicht ? )
  if
    drop drop 7     ( BEL auf Stack legen )
  else
    1+ adr @ c!     ( Count-Byte erhhen )
    adr @ dup c@ +  ( Zeichenadresse ermitteln )
    over swap c!    ( Zeichen ablegen )
  endif
;


: (get)  ( Zeichen von Tastatur fr Eingabestring holen )
  key       ( Taste holen )   ( ** Name gendert von GET )
  dup bs =  ( Backspace ? )
  if
    (bs)
  else
    dup 13 = ( Return-Taste ? )
    if
      drop 0    ( Null auf Stack legen )
    else
      dup 32 <  ( anderes Steuerzeichen ? )
      if
        drop 7  ( BEL auf Stack )
      else
        (app)
      endif
    endif
  endif
;


: expect ( adr count --> )
  (vector)  ( ** eingefgt )
  ml !             ( maximale Eingabelnge )
  0 over c!        ( Count auf Null setzen )
  adr !            ( Adresse in adr merken )
  begin
    (get) dup emit ( Zeichen holen und Echo )
    0=             ( bis CR gedrckt wird   )
  until
  adr @ dup c@ +   ( letztes Zeichen )
  1+ 0 swap c!     ( Nullbyte anhngen )
;


: query ( Zeile in TIB einlesen )
  tib @       ( Adresse holen )
  80 expect   ( bis 80 Zeichen lesen )
  0 in c!     ( IN auf Null setzen )
;



( *** formatierte Ausgabe *** )

: <# ( Start formatierte Ausgabe )
  pad hld !      ( Zeiger hld auf Scratchpad setzen )
  '0' leading !  ( fhrende Ziffer initialisieren   )
;


: hold ( Zeichen an Adresse in hld speichern )
  hld --    ( hld dekrementieren )
  hld @ c!  ( Zeichen ablegen    )
;


: # ( eine Ziffer ablegen )
   ddup or             ( auf Null prfen )
  if
    base c@ m/mod      ( Ziffer berechnen )
    rot '0' +          ( in ASCII-Format bringen )
    dup '9' >          ( Zahlenbasis grer 10 ? )
    if 7 + endif       ( ja, dann Korrektur )
  else
    leading @          ( Leerzeichen oder Null ausgeben )
    bl leading !       ( Leerzeichen als Default setzen )
  endif
   hold                ( Ziffer ablegen )
;


: #s ( restliche Ziffern ablegen )
  begin
    ddup or     ( Solange Zahl ungleich Null ... )
  while
    #           ( ... Ziffer ausgeben )
  repeat
;


: #> ( Ende formatierte Ausgabe, Zahl ist type-fertig )
  drop drop     ( Zahl entfernen )
  hld @ dup     ( hld-Adresse holen )
  pad - abs     ( Count-Byte erzeugen )
;


: sign ( Vorzeichen in formatierte Ausgabe einfgen )
  rot     ( Vorzeichen-Byte holen )
  0<      ( Vorzeichen negativ ? )
  if
    '-'   ( ja, dann Minuszeichen )
  else
    bl    ( nein, dann Leerzeichen )
  endif
  hold    ( Zeichen ausgeben )
;


: d. ( doppelt genaue Zahl ausgeben )
  swap over    ( Vorzeichen-Byte erzeugen )
  dabs         ( Absolutwert bilden )
  <# # #s      ( Zahl ausgeben )
  sign #>      ( Vorzeichen anfgen )
  type         ( ... und drucken )
;


: d.r ( doppelt genaue Zahl in spez. Feldbreite ausgeben )
  >r           ( Feldbreite merken )
  swap over    ( Vorzeichen-Byte erzeugen )
  dabs         ( Absolutwert bilden )
  r> 1-        ( Feldbreite holen )
  <# 1         ( Start Ausgabe )
  do # loop    ( Ausgabe entsprechend Feldbreite )
  sign #>      ( Vorzeichen anfgen )
  type         ( ... und drucken )
;


: . ( einfach genaue Zahl ausgeben )
  s->d   ( in doppelt genaue Zahl konvertieren )
  d.     ( ... und ausgeben )
;


: .r ( einfach genaue Zahl in spez. Feldbreite ausgeben )
  swap s->d   ( in doppelt genaue Zahl konvertieren )
  rot d.r     ( ... und ausgeben )
;


: u. ( vorzeichenlose Zahl ausgeben )
  0          ( auf doppelte Genauigkeit erweitern )
  <# # #s #> ( Konvertierung ohne Vorzeichen )
  type       ( ... und ausgeben )
;


: u.r ( vorzeichenlose Zahl in spez. Feldbreite ausgeben )
  0 swap       ( auf doppelte Genauigkeit erweitern )
  <# 1         ( Start Ausgabe )
  do # loop    ( Ausgabe entsprechend Feldbreite )
  #> type      ( ... und drucken )
;



( *** verschiedene Funktionen *** )

: ? ( Inhalt eines Speicherworts ausgeben )
  @ .
;


: char ( adr1 --> adr2 c )
  begin ( Zeichen holen, Leerzeichen berlesen )
    dup c@     ( Zeichen holen )
    dup bl =   ( Leerzeichen ? )
    while
      drop     ( Zeichen wegwerfen )
      1+       ( Adresse erhhen )
  repeat
;


: digit ( c --> ? flag )  ( Zeichen in Zahl konvertieren )
  '0' -          ( Zeichen mit '0' vergleichen )
  dup 0<         ( kleiner ? )
  if
    drop false   ( ja, keine Ziffer )
    exit         ( Abbruch )
  endif
  dup 9 >        ( Zeichen grer '9' ? )
  if
    7 - dup      ( ja, Hex-Korrektur )
    10 <         ( kleiner als 10 ? )
    if
      drop false ( ja, keine Ziffer )
      exit       ( Abbruch )
    endif
  endif
  dup base c@ <  ( mit base vergleichen )
  if
    true         ( ok, wenn kleiner )
  else
    drop false   ( sonst keine gltige Ziffer )
  endif
;


: (number) ( d1 adr1 --> d2 adr2 )
  begin
    in ++            ( Eingabe-Zhler erhhen )
    dup >r           ( Adresse merken )
    c@ digit         ( Zeichen holen und prfen )
    while            ( solange Ziffer )
      swap
      base @ u*      ( High-Word mit Base multiplizieren )
      drop rot                                  ( ** gendert )
      base @ u*      ( Low-Word mit Basis multiplizieren )
      d+             ( Ziffer aufaddieren )
      r> 1+          ( nchste Adresse )
  repeat
  r>
;


: number ( adr --> d ) ( String in doppeltgenaue Zahl konvertieren )
  1+ char         ( Leerzeichen berlesen )
  '-' =           ( negative Zahl ? )
  dup >r          ( merken )
  if
    1+ char drop  ( Minuszeichen berlesen )
  endif
  0 dup ddup      ( 4 Nullen auf Stack )
  dpl !           ( Dezimalpunkt lschen )
  in ! rot        ( Eingabezhler lschen )
  (number)        ( konvertieren )
  dup c@          ( Zeichen holen )
  '.' =           ( Dezimalpunkt ? )
  if
    in @ dpl !    ( Position merken )
    1+            ( nchste Adresse ) ( ** gendert )
    (number)      ( weiter konvertieren )
  endif
  drop            ( Adresse entfernen )
  r>              ( Vorzeichen holen )
  if dminus endif ( evtl. Vorzeichenwechsel )
;


: hex ( Zahlenbasis 16 setzen )
  16 base !
;


: decimal ( Zahlenbasis 10 setzen )
  10 base !
;


: .s ( Stack ausdrucken ) ( ** gendert )
  cr
  base @       ( Zahlenbasis sichern )
  decimal      ( Ausgabe dezimal )
  sp@          ( Adresse holen )
  2+ dup
  s0 @ 2- =    ( Stack leer ? )
  if
    ." STACK EMPTY"
  else
    ." TOS>"
    begin
      dup ? space  ( ausgeben )
      2+ dup       ( nchste Adresse )
      s0 @ 4 - >   ( fertig ? )
    until
    ." <BOTTOM"
  endif
  drop cr
  base !  ( Base restaurieren )
;


: min ( n1 n2 --> min )
  ddup >
  if swap endif
  drop
;


: max ( n1 n2 --> max )
  ddup <
  if swap endif
  drop
;



(save.voc)  ( Vocabulary Forth sichern )



( *** main *** )

: main
  7 emit ." Hello World"
;