program SForth;

{ S-Forth-Compiler Vers. 3.1 }

{ (c) 1987, 1988 Volker Everts & Erwin Schomburg }

{
** SForth ist ein Software-Projekt von
** DOS International
**
** Dokumentation siehe DOS Extra 2, 1988
** Die ursprngliche Version des Compilers
** wurde in den DOS-Ausgaben 9/87 bis 11/87
** verffentlicht.
**
** Seither wurde der SForth-Compiler unter
** Mitwirkung der Leser von DOS Inter-
** national stark modifiziert, verbessert
** und ausgebaut. Diese neueste Version 3.1
** des Compilers ist in erster Linie
** Erwin Schomburg zu verdanken, der durch
** viele Vorschlge und Anregungen ent-
** scheidenden Einflu auf die Weiterent-
** wicklung des SForth-Systems hatte.
}


const
  VERSION   = '3.1'; { Versionsnummer  }
  MAXSTACK  = 100;   { Stack-Ebenen    }
  EMPTY     = '';    { leerer String   }
  SPACE     = #32;   { Leerzeichen     }
  BEL       = #7;    { akust. Signal   }
  CR        = #13;   { Wagenrcklauf   }
  UNDEF     = 0;     { undefiniert     }
  KONSTANTE = 1;     { Forth-Konstante }
  VARIABLE  = 2;     { Forth-Variable  }
  MAKRO     = 3;     { Forth-Makro     }
  PROZEDUR  = 4;     { Forth-Prozedur  }
  SYSTEM    = 5;     { Systembefehl    }



type
  Memory = array[256..$7FFF] of byte; { 32 KByte }
  WortTyp = string[16];
  ZeilenTyp = string[127];
  HexStr = string[4];
  StackEintrag = record
                   Wert: integer;
                   Typ: byte;
                 end;
  StackTyp = array[0..MAXSTACK] of StackEintrag;



var
  infile: text;         { Forth-Quelltext   }
  outfile: file;        { compilierter Code }
  Zeile, LZ: ZeilenTyp; { Forth-Textzeile   }

  Wort,             { Forth-Wort }
  Name,             { Dateiname  }
  Ext,              { Extension  }
  TextFile,         { Quelldatei }
  VocName,          { Vocabulary }
  Merker,
  Merker2: WortTyp; { Merker     }

  sinit,            { Merker fr Stack-Initialisierung }
  addr,             { Hilfsadresse }
  maxallot,         { max. Speicher frei }
  pc,               { Programmzhler     }
  sp,               { Stackpointer       }
  Nummer,           { Zeilennummer       }
  Anfang,           { Anfang Dictionary  }
  Ende,             { Ende Dictionary    }
  Zeiger,           { auf aktuelles Wort }
  VocAnfang,        { Start Vocabulary   }
  ofcnt,            { OF-Zhler          }
  i,                { Zhlvariable       }
  mjmp: integer;    { Sprungadresse Main }

  M: Memory;       { Speicherbereich    }
  S: StackTyp;     { Kontroll-Stack     }

  main,            { Hauptprogramm      }
  dup,             { fr Optimierung    }
  sys,             { System-Befehl      }
  caseflag,
  gefunden,        { fr Wortsuche      }
  ListFlag,        { Listing ein/aus    }
  DumpFlag,        { Dump ein/aus       }
  NoCodeFlag,      { speichern ein/aus  }
  pcFlag: boolean; { pc zeigen ein/aus  }



function hex(n, l: integer): HexStr;

{ n in l-stellige Hexzahl wandeln }

var
  i, z: integer;
  s: HexStr;

begin
  s := EMPTY;
  for i := 1 to l do
    begin
      z := n and 15; { Ziffer bilden }
      if z > 9 then z := z + 7;
      s := chr(z + 48) + s;
      n := n shr 4; { Division durch 16 }
    end;
  hex := s;
end;



procedure dump;

{ Dictionary als Dump ausgeben }

var
  p, z: integer;
  cnt: integer;
  s: WortTyp;

begin
  p := Anfang;
  writeln;
  writeln('***** Start Dump *****');
  repeat
    cnt:=1;
    writeln;
    move(M[p], z, 2);
    if not (z=0) then
    begin
    write(hex(z, 4), SPACE);
    move(M[p+2], s, 17);
    write(s);
    for i:=length(s) to 13 do
    write(SPACE);
    p := p + 3 + length(s);
    case m[p] of
    1:write('KONSTANTE ');
    2:write('VARIABLE  ');
    3:write('MACRO     ');
    4:write('PROZEDUR  ');
    5:write('SYSTEM    ');
    else write('HUH?      ');
    end;
    write(M[p], SPACE);
    p := p + 1;
    while p < z do
      begin
        if (cnt>1) and ((cnt mod 16) = 1) then
        begin
        writeln;
        for i:=1 to 31 do
        write(SPACE);
        end;
        write(hex(M[p],2), SPACE);
        cnt:=succ(cnt);
        p := p + 1;
      end;
  end;
  until z=0;
  writeln;
  write('*** Ausgabe des fehlerhaften Restes ***');
  cnt:=1;
  for i := p to pc-1 do
  begin
  if (cnt mod 16)=1 then writeln;
  write(hex(M[i],2), SPACE);
  cnt:=succ(cnt);
  end;
  writeln;
end;



procedure error(nr: byte);

{ Fehlerbehandlung }

var msg: string[80];

begin
  case nr of
     1: msg := 'Syntax-Fehler';
     2: msg := 'Name erwartet';
     3: msg := 'doppelter Name';
     4: msg := 'undefinierter Name';
     5: msg := 'Proc-Aufrufe nicht erlaubt';
     6: msg := 'System-Befehle nicht erlaubt';
     7: msg := 'Zahlenwert erwartet';
     8: msg := 'Stack-berlauf';
     9: msg := 'Stack ist leer';
    10: msg := 'ENDIF oder ELSE ohne IF';
    11: msg := 'SHORT-Option unmglich';
    12: msg := 'unerwartetes Zeilenende';
    13: msg := 'fehlendes BEGIN';
    14: msg := 'fehlendes WHILE';
    15: msg := 'LOOP ohne DO';
    16: msg := 'Zeichenkonstante erwartet';
    17: msg := 'SF <source> [-L(isting) -P(c count) -D(ump) -N(o COM-File -M(axallot)]';
    18: msg := 'illegaler Parameter in Kommandozeile';
    19: msg := 'Dateifehler Quelltext';
    20: msg := 'IF ohne ENDIF';
    21: msg := 'BEGIN ohne UNTIL oder AGAIN';
    22: msg := 'WHILE ohne REPEAT';
    23: msg := 'DO ohne LOOP';
    24: msg := 'kein Vocabulary definiert';
    25: msg := 'Dateifehler bei INCLUDE';
    26: msg := 'falsche INCLUDE-Adresse';
    27: msg := ': MAIN Definition fehlt';
    28: msg := 'CASE-Schachtelung nicht erlaubt';
    else msg := 'unbekannter Fehler';
  end;
  writeln;
  if not ListFlag then writeln(LZ);
  writeln(BEL, '+++++ ', msg);
  if DumpFlag and not (Nr in [17, 18, 19]) then
    begin
      dump;
      writeln;
      writeln(LZ);
      writeln(msg);
    end;
  halt;
end;




procedure init;

{ Compiler initialisieren }

var
  p, i: byte;
  option: string[2];
  ch: char;

begin
  LZ := EMPTY;
  if ParamCount = 0 then error(17);

  { Dateinamen holen }

  TextFile := ParamStr(1);
  p := pos('.', TextFile);
  if p = 0 then
    begin
      Name := TextFile;
      Ext := '.FTH';
      TextFile := TextFile + Ext;
    end
  else
    begin
      Name := copy(TextFile, 1, p-1);
      Ext := copy(TextFile, p, 80);
    end;

  { Options-Voreinstellungen }

  ListFlag := false;
  DumpFlag := false;
  NoCodeFlag := false;
  pcFlag := false;
  maxallot:=0;

  { Optionen auswerten }

  if ParamCount > 1 then
    for i := 2 to ParamCount do
      begin
        option := ParamStr(i);
        if option[1] in ['/', '-'] then
          begin
            ch := UpCase(option[2]);
            case ch of
              'L': ListFlag := true;
              'D': DumpFlag := true;
              'N': NoCodeFlag := true;
              'P': pcFlag := true;
              'M': begin
                     Merker:=copy(paramstr(i), 3, 80);
                     val(Merker, maxallot, addr);
                     if (addr <> 0) then error(7);
                   end;
              else error(18);
            end; { case }
          end { if }
        else if option[1] <> '>' then error(18);
      end; { for }

  { Quelltextdatei ffnen }

  assign(infile, TextFile);
  {$I-} reset(infile); {$I+}
  if IOresult <> 0 then error(19);

  { verschiedene Einstellungen }

  Zeile := EMPTY;
  Merker := EMPTY;
  Wort := SPACE;
  main := false;
  dup := false;
  sp := 0;
  Nummer := 0;
  pc := 256;
  VocName := EMPTY;
  ofcnt:=0;
  caseflag:=false;
  VocAnfang := 0;
end;



function HoleWort: WortTyp;

{ Ein Wort aus Quelltext holen }

label leave;

var
  p: byte;
  w: WortTyp;

begin
  repeat
    if Zeile = EMPTY then
      if eof(infile) then
        begin
          HoleWort := EMPTY;
          goto leave;
        end
      else
        begin
          readln(infile, Zeile);
          LZ := Zeile;
          if pcFlag then write(hex(pc,4), SPACE, SPACE);
          Nummer := succ(Nummer);
          if ListFlag then writeln(Nummer:5, '   ', Zeile)
          else write(Nummer:5, CR);
        end;
    while copy(Zeile, 1, 1) = SPACE do delete(Zeile, 1, 1);
  until zeile <> EMPTY;
  p := pos(SPACE, Zeile);
  if p = 0 then
    begin
      w := Zeile;
      Zeile := EMPTY;
    end
  else
    begin
      w := copy(Zeile, 1, p-1);
      Zeile := copy(Zeile, p+1, 255);
    end;
  Merker2 := w;
  for p := 1 to length(w) do w[p] := UpCase(w[p]);
  dup := Merker = 'DUP';
  Merker := w;
  HoleWort := w;
  Leave:
end;



function HoleZeichen: char;

{ Ein einzelnes Zeichen aus Quelltext holen }

begin
  if Zeile = EMPTY then error(12);
  HoleZeichen := Zeile[1];
  Zeile := copy(Zeile, 2, 255);
end;



procedure ob(b: Byte);

{ Ein Byte in Dictionary ablegen }

begin
  M[pc] := b;
  pc := pc + 1;
end;



procedure ow(w: integer);

{ Ein Wort in Dictionary ablegen }

begin
  move(w, M[pc], 2);
  pc := pc + 2;
end;



procedure ot(adr, w: integer);

{ Ein Wort an spezifizierter Adresse ablegen }

begin
  move(w, M[adr], 2);
end;



procedure os(s: WortTyp);

{ String im Dictionary ablegen }

begin
  for i := 0 to length(s) do ob(ord(s[i]));
end;



procedure TueKommentar;

{ Kommentar berlesen }

var z: char;

begin
  repeat
    z := HoleZeichen;
  until (z = ')') or (Zeile = EMPTY);
end;



procedure TueCode;

{ Inline-Code auswerten }

label 99;

var
  w: WortTyp;
  disp, Fehler: integer;

begin
  repeat
    w := HoleWort;
    if w = EMPTY then error(7);
    if w = ']' then goto 99;
    val(w, disp, Fehler);
    if Fehler <> 0 then error(7);
    if disp > 255 then ow(disp) else ob(disp);
    99:
  until w = ']';
end;



procedure push(p: integer; Flag: Byte);

{ Adresse und Flag auf Stack ablegen }

begin
  S[sp].Wert := p;
  S[sp].Typ := Flag;
  sp := sp + 1;
  if sp > MaxStack then error(8);
end;



function pop(Flag: Byte): integer;

{ Adresse vom Stack holen, Flag prfen }

begin
  sp := sp - 1;
  if sp < 0 then error(9);
  if S[sp].Typ = Flag then pop := S[sp].Wert
  else
    case Flag of
      1: error(10);
      2: error(13);
      3: error(14);
      4: error(15);
      else error(0);
    end;
end;



function near(Quelle, Ziel: integer): integer;

{ Near-Sprungdistanz berechnen }

var disp: integer;

begin
  near := Ziel - Quelle - 2;
end;




procedure Suche(Name: WortTyp; var gef: boolean);

{ Namen in Dictionary suchen }

label 99;

var
  z: integer;
  n: WortTyp;

begin
  z := Anfang;
  n := EMPTY;
  gef := false;
  while Name <> n do
    begin
      Zeiger := z;
      move(M[Zeiger], z, 2);
      if z = 0 then goto 99;
      move(M[Zeiger+2], n, 17);
    end;
  99: gef := Name = n;
end;



function FindeVar(w: worttyp):integer;
begin
      suche(w, gefunden);
      if not(gefunden) then error(4);
      if M[Zeiger+length(w)+3] <> Variable then error(4);
      findevar := Zeiger + length(w) + 12;
end;



procedure TueName;

{ Name holen und berprfen, Header bauen }

var Name: WortTyp;

begin
  Name := HoleWort;
  if Name = EMPTY then error(2);
  Suche(Name, gefunden);
  if gefunden then error(3);
  os(Name);
  main := Name = 'MAIN';
end;



procedure pushregs;
begin
  ob($06);   { push es }
  ob($1e);   { push ds }
  ob($57);   { push di }
  ob($56);   { push si }
  ob($55);   { push bp }
end;



procedure popregs;
begin
  ob($5d);    { pop bp }
  ob($5e);    { pop si }
  ob($5f);    { pop di }
  ob($1f);    { pop ds }
  ob($07);    { pop es }
end;



procedure getvars;
begin
  ob($a1); ow(findevar('AX'));           { mov ax,[AX] }
  ob($8b); ob($1e); ow(findevar('BX'));  { mov bx,[BX] }
  ob($8b); ob($0e); ow(findevar('CX'));  { mov cx,[CX] }
  ob($8b); ob($16); ow(findevar('DX'));  { mov dx,[DX] }
end;



procedure setvars;
begin
  ob($89); ob($16); ow(findevar('DX'));  { mov [DX],dx }
  ob($89); ob($0e); ow(findevar('CX'));  { mov [CX],cx }
  ob($89); ob($1e); ow(findevar('BX'));  { mov [BX],bx }
  ob($a3); ow(findevar('AX'));           { mov [AX],ax }
end;



procedure TueSystem(w: WortTyp);

{ SYSTEM-Worte compilieren }

label 99;

var
  z: char;
  disp, Fehler, adr: integer;

begin
  sys := true;

  { ; }

  if w = ';' then
    begin
      if main then
        begin
          ob($B4); ob($00);    { mov ah,0 }
          ob($CD); ob($21);    { int 21H  }
          ob($90);             { nop      }
          addr := pc;
          move(addr, M[findevar('TIB')], 2);
          addr := pc + 80;
          move(addr, M[findevar('DP')], 2);
          addr := addr + maxallot + $44 + 80 + $80;
          move(addr, M[findevar('S0')], 2);
        end
      else ob($C3);            { ret }
      if sp > 0 then
        case S[sp-1].typ of
          1: error(20);
          2: error(21);
          3: error(22);
          4: error(23);
          else error(0);
        end;
    end

  { ( }

  else if w = '(' then TueKommentar

  { [  }

  else if w = '[' then TueCode

  { IF }

  else if w = 'IF' then
    begin
      if dup then pc := pc - 4;
      ob($85); ob($C0);          { test ax,ax }
      if not dup then ob($AD);   { lodsw      }
      ob($75); ob(03);           { jnz  +3    }
      ob($e9);                   { jmp   disp  }
      push(pc, 1); ow(0);
    end

  { ENDIF }

  else if w = 'ENDIF' then
    begin
      adr := pop(1);
      M[adr] := lo(near(adr, pc));
      M[adr+1] := hi(near(adr, pc));
    end

  { ELSE, ENDOF, ;; }

  else if (w = 'ELSE') or (w = 'ENDOF') or (w = ';;') then
    begin
      adr := pop(1);
      M[adr] := lo(near(adr, pc+3));
      M[adr+1] := hi(near(adr, pc+3));
      ob($E9);                    { jmp  disp }
      push(pc, 1); ow(0);
    end

  { CASE }

  else if w = 'CASE' then
    begin
      if caseflag then error(28);
      caseflag:=true;
    end

  { OF }

  else if w = 'OF' then
    begin
      ofcnt := succ(ofcnt);
      ob($2b); ob($04);   { sub ax,[si] }
      ob($ad);            { lodsw       }
      ob($74);ob(03);     { jz  +3      }
      ob($e9);            { jmp disp    }
      push(pc, 1); ow(0);
      ob($ad);            { lodsw       }
    end

  { ENDCASE }

  else if w = 'ENDCASE' then
    begin
      ob($ad);            { lodsw }
      for i:=1 to ofcnt do
        begin
          adr := pop(1);
          M[adr] := lo(near(adr, pc));
          M[adr+1] := hi(near(adr, pc));
        end;
      ofcnt:=0;
      caseflag:=false;
    end

  { FIND }

  else if w = 'FIND' then
    begin
      w := HoleWort;
      suche(w, gefunden);
      if not(gefunden) then error(4);
      adr := Zeiger + length(w) + 4;
      ob($4e);           { dec si      }
      ob($4e);           { dec si      }
      ob($89); ob($04);  { mov [si],ax }
      ob($b8); ow(adr);  { mov ax,adr  }
    end

  { INT }

  else if w = 'INT' then
    begin
      ob($a2); ow(pc+19);  { mov [adr],al }
      ob($51);             { push cx      }
      getvars;
      ob($CD);ob(0);       { int ...  }
      setvars;
      ob($59);             { pop cx   }
      ob($ad);             { lodsw    }
    end

  { BEGININT }

  else if w = 'BEGININT' then
    begin
      pushregs;
      ob($0e);    { push cs }
      ob($1f);    { pop ds  }
      setvars;
      ob($8b); ob($36); ow(findevar('S0')); { mov si,[s0] }
      ob($89); ob($26); ow(findevar('R0')); { mov [r0],sp }
    end

  { ENDINT }

  else if w = 'ENDINT' then
    begin
      getvars;
      popregs;
      ob($cf); { iret }
    end

  { EXITTO }

   else if w = 'EXITTO' then
     begin
       ob($fa);                  { cli           }
       ob($89);ob($c3);          { mov bx,ax     }
       ob($8b);ob($07);          { mov ax,[bx]   }
       ob($a3);ow(pc + 25);      { mov [disp],ax }
       getvars;
       popregs;
       ob($2e); ob($ff); ob($2e); ow(0); { jmp far ... }
     end

  { CALLTO }

   else if w = 'CALLTO' then
     begin
       ob($fa);              { cli           }
       ob($89); ob($c3);     { mov bx,ax     }
       ob($8b); ob($07);     { mov ax,[bx]   }
       ob($a3); ow(pc + 26); { mov [disp],ax }
       getvars;
       popregs;
       ob($9c);              { pushf         }
       ob($2e); ob($ff); ob($1e); ow(0); { call far ... }
       ob($fb);              { sti           }
       pushregs;
       ob($0e);              { push cs       }
       ob($1f);              { pop ds        }
       ob($8b); ob($36);
       ow(findevar('S0'));   { mov si,[s0]   }
       setvars;
     end

  { BEGIN }

  else if w = 'BEGIN' then
    begin
      push(pc, 2);
    end

  { UNTIL }

  else if w = 'UNTIL' then
    begin
      if dup then pc := pc - 4;
      ob($85); ob($C0);           { test ax,ax }
      if not dup then ob($AD);    { lodsw      }
      adr := pop(2);
      ob($75); ob(03);            { jnz  +3    }
      ob($e9); ow(near(pc,adr));  { jmp  disp  }
    end

  { WHILE }

  else if w = 'WHILE' then
    begin
      if dup then pc := pc - 4;
      ob($85); ob($C0);        { test ax,ax }
      if not dup then ob($AD); { lodsw      }
      ob($75); ob(03);         { jnz  +3    }
      ob($e9);                 { jmp  disp  }
      push(pc, 3); ow(0);
    end

  { REPEAT }

  else if w = 'REPEAT' then
    begin
      adr := pop(3);
      M[adr] := lo(near(adr, pc+3));
      M[adr+1] := hi(near(adr, pc+3));
      adr := pop(2);
      ob($E9); ow(near(pc,adr)); { jmp  disp }
    end

  { AGAIN }

  else if w = 'AGAIN' then
    begin
      adr := pop(2);
      ob($E9); ow(near(adr,pc)); { jmp  disp }
    end

  { DO }

  else if w = 'DO' then
    begin
      ob($57);                { push di    }
      ob($51);                { push cx    }
      ob($89); ob($C1);       { mov  cx,ax }
      ob($AD);                { lodsw      }
      ob($89); ob($C7);       { mov  di,ax }
      ob($AD);                { lodsw      }
      push(pc, 4);
    end

  { LOOP }

  else if w = 'LOOP' then
    begin
      ob($41);                   { inc cx    }
      ob($39); ob($F9);          { cmp cx,di }
      adr := pop(4);
      ob($7f); ob(03);           { jg  +3    }
      ob($e9); ow(near(pc,adr)); { jmp adr   }
      ob($59);                   { pop cx    }
      ob($5F);                   { pop di    }
    end

  { +LOOP }

  else if w = '+LOOP' then
    begin
      if dup then pc := pc - 4;
      ob($03); ob($C8);           { add cx,ax }
      if not dup then ob($AD);    { lodsw     }
      ob($39); ob($F9);           { cmp cx,di }
      adr := pop(4);
      ob($7f); ob(03);            { jg  +3    }
      ob($e9); ow(near(pc,adr));  { jmp adr   }
      ob($59);                    { pop cx    }
      ob($5F);                    { pop di    }
    end

  { -LOOP }

  else if w = '-LOOP' then
    begin
      if dup then pc := pc - 4;
      ob($29); ob($C1);           { sub cx,ax }
      if not dup then ob($AD);    { lodsw     }
      ob($39); ob($F9);           { cmp cx,di }
      adr := pop(4);
      ob($7c); ob(03);            { jl  +3    }
      ob($e9); ow(near(pc,adr));  { jmp adr   }
      ob($59);                    { pop cx    }
      ob($5F);                    { pop di    }
    end

  { /LOOP }

  else if w = '/LOOP' then
    begin
      if dup then pc := pc - 4;
      adr := pop(4);
      ob($03); ob($C8);           { add  cx,ax }
      ob($85); ob($C0);           { test ax,ax }
      if not dup then ob($AD);    { lodsw      }
      ob($79); ob($07);           { jns  disp  }
      ob($39); ob($F9);           { cmp  cx,di }
      ob($7c); ob($0a);           { jl   +10   }
      ob($e9); ow(near(pc,adr));  { jmp  adr   }
      ob($39); ob($F9);           { cmp  cx,di }
      ob($7f); ob(03);            { jg   +3    }
      ob($e9); ow(near(pc,adr));  { jmp  adr   }
      ob($59);                    { pop  cx    }
      ob($5F);                    { pop  di    }
    end

  { (") }

  else if w = '(")' then
    begin
      adr := pc;
      ob(0);
      z := HoleZeichen;
      while (z <> '"') and (z <> EMPTY) do
        begin
          ob(ord(z)); z := HoleZeichen;
          M[adr] := M[adr] + 1;
        end;
    end

  else sys := false;
end;



procedure TueProc;

{ Colon-Definition (Prozedur) compilieren }

label ok;

var
  w: WortTyp;
  disp, Fehler, adr, typ: integer;

begin
  TueName;        { Header bauen }
  ob(PROZEDUR);
  if main then ot(mjmp, pc - mjmp -2);

  repeat
    w := HoleWort;
    if w = EMPTY then error(4);

    { Char-Konstante ? }

    if w[1] = #39 { ' } then
      begin
        if length(w) <> 3 then error(16);
        if w[3] <> #39 then error(16);
        disp := ord(Merker2[2]);
        ob($4E);           { dec si      }
        ob($4E);           { dec si      }
        ob($89); ob($04);  { mov [si],ax }
        ob($B8); ow(disp); { mov ax,disp }
        goto ok;
      end;

    { Numerische Konstante ? }

    if (length(w) = 1) and (w[1] in ['+', '-', '0', '1', '2']) then Fehler := 1
    else val(w, disp, Fehler);
    if Fehler = 0 then
      begin
        ob($4E);           { dec si      }
        ob($4E);           { dec si      }
        ob($89); ob($04);  { mov [si],ax }
        ob($B8); ow(disp); { mov ax,disp }
        goto ok;
      end;

    { System-Wort ? }

    TueSystem(w);
    if sys then goto ok;

    { in Dictionary suchen }

    suche(w, gefunden);
    if not(gefunden) then error(4);
    adr := Zeiger + length(w) + 4;
    typ := M[adr-1];

    { Makro ? }

    if typ = MAKRO then
      begin
        move(M[Zeiger], disp, 2);
        while adr < disp do
          begin
            ob(M[adr]);
            adr := adr + 1;
          end;
        goto ok;
      end;

    { sonst Konstante, Variable oder Prozedur }

    disp := adr - pc - 3;
    ob($E8); ow(disp); { call disp }
    if w = '."' then TueSystem('(")');
    if w = '"' then TueSystem('(")');
    ok:
  until w = ';';
  ot(Ende, pc);
  Ende := pc;
  ow(0);
end;



procedure TueMacro;

{ Makro-Definition compilieren }

label ok;

var
  w: WortTyp;
  disp, Fehler, adr, typ: integer;

begin
  TueName;    { Header bauen }
  ob(MAKRO);

  repeat
    w := HoleWort;
    if w = EMPTY then error(4);

    { Char-Konstante ? }

    if w[1] = #39 { ' } then
      begin
        if length(w) <> 3 then error(16);
        if w[3] <> #39 then error(16);
        disp := ord(Merker2[2]);
        ob($4E);           { dec si      }
        ob($4E);           { dec si      }
        ob($89); ob($04);  { mov [si],ax }
        ob($B8); ow(disp); { mov ax,disp }
        goto ok;
      end;

    { Numerische Konstante ? }

    val(w, disp, Fehler);
    if Fehler = 0 then
      begin
        ob($4E);           { dec si      }
        ob($4E);           { dec si      }
        ob($89); ob($04);  { mov [si],ax }
        ob($B8); ow(disp); { mov ax,disp }
        goto ok;
      end;

    if w = ';' then goto ok  { Makro-Ende }
    else if w = '(' then TueKommentar
    else if w = '[' then TueCode
    else
      begin { Wort in Dictionary suchen }
        suche(w, gefunden);
        if not(gefunden) then error(4);
        adr := Zeiger + length(w) + 4;
        typ := M[adr-1];
        if typ = MAKRO then
          begin
            move(M[Zeiger], disp, 2);
            while adr < disp do
              begin
                ob(M[adr]);
                adr := adr + 1;
              end;
          end
        else error(5); { Konstanten, Variable und }
      end;             { Prozeduren nicht erlaubt }
    ok:
  until w = ';';
  ot(Ende, pc);
  Ende := pc;
  ow(0);
end;



procedure TueVariable;

{ Variablen-Definition compilieren }

var
  w: WortTyp;
  n, Fehler: integer;

begin
  TueName;        { Header bauen }
  ob(VARIABLE);
  w := HoleWort;
  if w = EMPTY then error(7);

  { Char-Konstante ? }

  if w[1] = #39 { ' } then
    begin
      if length(w) <> 3 then error(16);
      if w[3] <> #39 then error(16);
      n := ord(Merker2[2]);
    end
  else     { numerische Konstante }
    begin
      val(w, n, Fehler);
      if Fehler <> 0 then error(7);
    end;
  ob($4E);           { dec si      }
  ob($4E);           { dec si      }
  ob($89); ob($04);  { mov [si],ax }
  ob($B8); ow(pc+3); { mov ax,n    }
  ob($C3);           { ret         }
  ow(n);
  ot(Ende, pc);
  Ende := pc;
  ow(0);
end;



procedure TueKonstante;

{ Konstanten-Definition compilieren }

var
  w: WortTyp;
  n, Fehler: integer;

begin
  TueName;         { Header bauen }
  ob(KONSTANTE);
  w := HoleWort;
  if w = EMPTY then error(7);

  { Char-Konstante ? }

  if w[1] = #39 { ' } then
    begin
      if length(w) <> 3 then error(16);
      if w[3] <> #39 then error(16);
      n := ord(Merker2[2]);
    end
  else     { numerische Konstante }
    begin
      val(w, n, Fehler);
      if Fehler <> 0 then error(7);
    end;
  ob($4E);          { dec si      }
  ob($4E);          { dec si      }
  ob($89); ob($04); { mov [si],ax }
  ob($B8); ow(n);   { mov ax,n    }
  ob($C3);          { ret         }
  ot(Ende, pc);
  Ende := pc;
  ow(0);
end;



procedure TueVoc;

{ Vocabulary merken }

begin
  VocName := HoleWort;
  if VocName = EMPTY then error(2);
  VocAnfang := Ende;
end;



procedure SichereVoc;

{ aktuelles Vocabulary sichern }

var
  v: file;
  i: integer;
  b1, b2: byte;

begin
  if VocAnfang = 0 then error(24);
  assign(v, VocName + '.VOC');
  rewrite(v,1);
  writeln;
  writeln('Sichern ', VocName, '.VOC ...');
  b1 := lo(VocAnfang);
  b2 := hi(VocAnfang);
  BlockWrite(v, b1, 1);
  BlockWrite(v, b2, 1);
  BlockWrite(v, M[VocAnfang], pc - VocAnfang);
  close(v);
  writeln('Fertig');
  writeln;
end;



procedure Include;

{ Vocabulary einbinden }

var
  Name: WortTyp;
  v: file;
  i: integer;
  b1, b2: byte;

begin
  Name := HoleWort;
  if Name = EMPTY then error(2);
  Name := Name + '.VOC';
  writeln;
  writeln('Lesen ', Name, ' ...');
  assign(v, Name);
  {$I-} reset(v,1); {$I+}
  if IOresult <> 0 then error(25);
  BlockRead(v, b1, 1);
  BlockRead(v, b2, 1);
  i := b2 shl 8 + b1;
  if i <> Ende then error(26);
  pc := pc - 2;
  BlockRead(v, M[pc], FileSize(v) - 2);
  pc := pc + FileSize(v) - 2;
  Ende := pc - 2;
  close(v);
  writeln('Fertig');
  writeln;
end;


procedure compile(w: WortTyp);

{ Wort compilieren }

var i: byte;

begin
  if w = ':' then TueProc
  else if w = 'MACRO' then TueMacro
  else if w = 'VARIABLE' then TueVariable
  else if w = 'CONSTANT' then TueKonstante
  else if w = '(' then TueKommentar
  else if w = 'VOCABULARY' then TueVoc
  else if w = 'INCLUDE' then Include
  else if w = '(DUMP)' then dump
  else if w = '(LIST.ON)' then ListFlag := true
  else if w = '(LIST.OFF)' then ListFlag := false
  else if w = '(SAVE.VOC)' then SichereVoc
  else error(1);
end;



{ Hauptprogramm S-Forth }


begin
  ClrScr;
  writeln;
  writeln('*** S-Forth Compiler Vers. ', VERSION, ' ***');
  writeln;
  writeln('/// DOS International');
  writeln;
  init;
  writeln('Start Compilation');
  writeln;
  begin
    ob($8B); ob($36);
    sinit := pc;
    ow(0);                     { mov si,[S0]   }
    ob($33); ob($C0);          { xor ax,ax     }
    ob($E9); ow($0000);        { jmp main      }
  end;
  Anfang := pc;
  Ende := pc;
  ow(0);
  mjmp := pc-4;
  Wort := HoleWort;
  while (Wort <> EMPTY) and (not main) do
    begin
      compile(Wort);
      Wort := HoleWort;
    end;
  ot(sinit, findevar('S0'));
  if NoCodeFlag = false then
    begin
      if not main then error(27);
      writeln; writeln;
      writeln('Speichern ', Name + '.com');
      assign(outfile, Name + '.com');
      rewrite(outfile, pc-256);
      BlockWrite(outfile, M[256], 1);
      close(outfile);
    end;
  writeln('Fertig');
end.
